﻿using System;
using System.Net.NetworkInformation;
using System.Threading;

namespace CountDownTimerEvent
{
    class Program
    {
        private static void Main(string[] args)
        {
            MainDelegate();
            //MainEvent();
            //MainEventAddRemove();
            //MainRaisingExceptionRequiresProtection();
        }

        static void MainDelegate()
        {
            CountDownDelegate countDown = new CountDownDelegate(5);
            countDown.Tick +=
                currentSecond => Console.WriteLine(currentSecond);
            // MultiCastDelegate:
            // Notice the difference between = and +=
            // Using = you "hijack" the event
            countDown.Tick =
                currSec => Console.WriteLine("Another: " + currSec);
            countDown.Finished +=
                () => Console.WriteLine("Done");
            countDown.Start();

            Thread.Sleep(2000);
            countDown.Tick(400);
            // An extra Tick, from OUTSIDE the CountDown class!!
        }

        static void MainEvent()
        {
            CountDownEvent countDown = new CountDownEvent(5);
            countDown.Tick +=
                currentSecond => Console.WriteLine(currentSecond);
            // Event (not delegate): // No "hijacking"
            countDown.Tick +=
                currSec => Console.WriteLine("Another: " + currSec);
            countDown.Finished +=
                () => Console.WriteLine("Done");
            countDown.Start();

            Thread.Sleep(2000);

            //countDown.Tick(400);
            // Event (not delegate): No extra events raised from outsiders
        }
        static void MainEventAddRemove()
        {
            CountDownEvent countDown = new CountDownEvent(5);
            countDown.Tick +=
                currentSecond => Console.WriteLine("First: " + currentSecond);

            // Named delegate object
            Action<uint> countDownOnTick = currSec => Console.WriteLine("Another: " + currSec);
            countDown.Tick += countDownOnTick;
            countDown.Finished +=
                () => Console.WriteLine("Done");
            countDown.Start();

            Thread.Sleep(2000);

            // Remove delegate object using -=
            countDown.Tick -= countDownOnTick;
        }

        private static event Action MyEvent;

        // De Smet, page 857
        static void MainRaisingExceptionRequiresProtection()
        {
            new Thread(() =>
            {
                while (true)
                {
                    Action a = () => Console.WriteLine("Hello");
                    MyEvent += a;
                    MyEvent -= a;
                }
            }).Start();
            while (true)
            {
                if (MyEvent != null)
                {
                    //Thread.Sleep(2);
                    MyEvent(); // raise the event
                }

                // Be careful, take a copy of the event object to avoid Null Reference
                //Action ev = MyEvent;
                //if (ev != null)
                //{
                //    //Thread.Sleep(2);
                //    ev(); // raise the event
                //}
            }
        }


    }
}
